@@ -4,6 +4,65 @@
 require 'pathname'
 module Rcov
 
+# Try to fix bugs in the REXML shipped with Ruby 1.8.6
+# They affect Mac OSX 10.5.1 users and motivates endless bug reports.
+begin 
+    require 'rexml/formatters/transitive'
+    require 'rexml/formatter/pretty'
+rescue LoadError
+end
+
+if RUBY_VERSION == "1.8.6" && defined? REXML::Formatters::Transitive
+    class REXML::Document
+        remove_method :write rescue nil
+        def write( output=$stdout, indent=-1, trans=false, ie_hack=false )
+            if xml_decl.encoding != "UTF-8" && !output.kind_of?(Output)
+                output = Output.new( output, xml_decl.encoding )
+            end
+            formatter = if indent > -1
+                #if trans
+                    REXML::Formatters::Transitive.new( indent )
+                #else
+                #    REXML::Formatters::Pretty.new( indent, ie_hack )
+                #end
+            else
+                REXML::Formatters::Default.new( ie_hack )
+            end
+            formatter.write( self, output )
+        end
+    end
+
+    class REXML::Formatters::Transitive
+        remove_method :write_element rescue nil
+        def write_element( node, output )
+            output << "<#{node.expanded_name}"
+
+            node.attributes.each_attribute do |attr|
+                output << " "
+                attr.write( output )
+            end unless node.attributes.empty?
+
+            if node.children.empty?
+                output << "/>" 
+            else
+                output << ">"
+                # If compact and all children are text, and if the formatted output
+                # is less than the specified width, then try to print everything on
+                # one line
+                skip = false
+                @level += @indentation
+                node.children.each { |child|
+                    write( child, output )
+                }
+                @level -= @indentation
+                output << "</#{node.expanded_name}>"
+            end
+            output << "\n"
+            output << ' '*@level
+        end
+    end
+end
+
 class Formatter # :nodoc:
     require 'pathname'
     ignore_files = [
@@ -226,7 +285,17 @@ class FullTextReport < Formatter # :nodoc:
             puts "=" * 80
             puts filename
             puts "=" * 80
-            SCRIPT_LINES__[filename].each_with_index do |line, i|
+            lines = SCRIPT_LINES__[filename]
+            unless lines
+                # try to get the source code from the global code coverage
+                # analyzer
+                re = /#{Regexp.escape(filename)}\z/
+                if $rcov_code_coverage_analyzer and 
+                    (data = $rcov_code_coverage_analyzer.data_matching(re))
+                    lines = data[0]
+                end
+            end
+            (lines || []).each_with_index do |line, i|
                 case @textmode
                 when :counts
                     puts "%-70s| %6d" % [line.chomp[0,70], fileinfo.counts[i]]
@@ -616,7 +685,7 @@ EOS
 
     DEFAULT_OPTS = {:color => false, :fsr => 30, :destdir => "coverage",
                     :callsites => false, :cross_references => false,
-                    :validator_links => true 
+                    :validator_links => true, :charset => nil
                    }
     def initialize(opts = {})
         options = DEFAULT_OPTS.clone.update(opts)
@@ -628,6 +697,7 @@ EOS
         @do_cross_references = options[:cross_references]
         @span_class_index = 0
         @show_validator_links = options[:validator_links]
+        @charset = options[:charset]
     end
 
     def execute
@@ -699,7 +769,7 @@ EOS
                              [f.code_coverage, "coverage_code"]].each do |value, css_class|
                                 value *= 100
                                 td_ { 
-                                    table_(:cellpadding => 0, :cellspacing => 0, :align => "right") { 
+                                    table_(:cellpadding => "0", :cellspacing => "0", :align => "right") { 
                                         tr_ { 
                                             td_ {
                                                  tt_(:class => css_class) { "%3.1f%%" % value } 
@@ -707,11 +777,11 @@ EOS
                                             }
                                             ivalue = value.round
                                             td_ {
-                                                table_(:class => "percent_graph", :cellpadding => 0,
-                                                   :cellspacing => 0, :width => 100) {
+                                                table_(:class => "percent_graph", :cellpadding => "0",
+                                                   :cellspacing => "0", :width => "100") {
                                                     tr_ {
-                                                        td_(:class => "covered", :width => ivalue)
-                                                        td_(:class => "uncovered", :width => (100-ivalue))
+                                                        td_(:class => "covered", :width => ivalue.to_s)
+                                                        td_(:class => "uncovered", :width => (100-ivalue).to_s)
                                                     }
                                                 }
                                             }
@@ -740,6 +810,10 @@ EOS
         title = default_title
         output = xhtml_ { html_ {
             head_ {
+                if @charset
+                    meta_("http-equiv".to_sym => "Content-Type",
+                          :content => "text/html;charset=#{@charset}")
+                end
                 title_{ title }
                 style_(:type => "text/css") { t_{ "body { background-color: #{default_color}; }" }  }
                 style_(:type => "text/css") { CSS_PROLOG }
@@ -763,7 +837,7 @@ EOS
                     p_ {
                         a_(:href => "http://validator.w3.org/check/referer") {
                             img_(:src => "http://www.w3.org/Icons/valid-xhtml11",
-                                 :alt => "Valid XHTML 1.1!", :height => 31, :width => 88)
+                                 :alt => "Valid XHTML 1.1!", :height => "31", :width => "88")
                         }
                         a_(:href => "http://jigsaw.w3.org/css-validator/check/referer") {
                             img_(:style => "border:0;width:88px;height:31px",
@@ -801,7 +875,7 @@ EOS
                     end_of_span = "</span>"
                 end
             end
-            result += %[<a name="line#{i+1}" />] + (format_line % (i+1)) + 
+            result += %[<a name="line#{i+1}"></a>] + (format_line % (i+1)) + 
                 " " + create_cross_refs(file.name, i+1, CGI.escapeHTML(line)) + "\n"
             last = spanclass
         end
@@ -877,6 +951,10 @@ EOS
         do_ctable = output_color_table?
         output = xhtml_ { html_ {
             head_ { 
+                if @charset
+                    meta_("http-equiv".to_sym => "Content-Type",
+                          :content => "text/html;charset=#{@charset}")
+                end
                 title_{ title } 
                 style_(:type => "text/css") { t_{ "body { background-color: #{default_color}; }" }  }
                 style_(:type => "text/css") { CSS_PROLOG }
@@ -910,7 +988,7 @@ EOS
                     p_ {
                         a_(:href => "http://validator.w3.org/check/referer") {
                             img_(:src => "http://www.w3.org/Icons/valid-xhtml10",
-                                 :alt => "Valid XHTML 1.0!", :height => 31, :width => 88)
+                                 :alt => "Valid XHTML 1.0!", :height => "31", :width => "88")
                         }
                         a_(:href => "http://jigsaw.w3.org/css-validator/check/referer") {
                             img_(:style => "border:0;width:88px;height:31px",
@@ -1080,8 +1158,8 @@ class RubyAnnotation < Formatter # :nodoc:
         return linetext unless @callsite_analyzer && @do_callsites
         ref_blocks = []
         _get_defsites(ref_blocks, filename, lineno, linetext, ">>") do |ref|
-            ref.file.sub!(%r!^./!, '')
             if ref.file
+                ref.file.sub!(%r!^./!, '')
                 where = "at #{mangle_filename(ref.file)}:#{ref.line}"
             else
                 where = "(C extension/core)"
